/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import java.util.Comparator;

/**
 * Provides access to the array characteristics.
 *
 * @author Leon van Zantvoort
 */
public interface ArrayCharacteristics<E> {

    /**
     * Characteristic value signifying that the elements of array are not null.
     */
    static final int NONNULL = 0x00000001;

    /**
     * Characteristic value signifying that the array is mutable. Elements can be replaced.
     */
    static final int MUTABLE = 0x00000002;

    /**
     * Characteristic value signifying that the array is capacity bounded. Elements can be added and removed.
     */
    static final int BOUNDED = 0x00000004;

    /**
     * Characteristic value signifying that the array is capacity unbounded. Elements can be added and removed at any
     * index.
     */
    static final int UNBOUNDED = 0x00000008;

    /**
     * Characteristic value signifying that the elements of the array are stored in a defined sort order.
     */
    static final int SORTED = 0x00000010;

    /**
     * Characteristic value signifying that for each pair of elements of the array {@code x, y}, {@code !x.equals(y)}.
     */
    static final int DISTINCT = 0x00000020;

    /**
     * Tells whether or not this array is backed by an array.
     *
     * <p>If this method returns {@code true} then array instance may safely be obtained through {@code array()}.</p>
     *
     * @return {@code true} if, and only if, this array is backed by an array.
     */
    boolean hasArray();

    /**
     * Tells whether or not this array is backed by a buffer.
     *
     * <p>If this method returns {@code true} then buffer instance may safely be obtained through {@code buffer()}.</p>
     *
     * @return {@code true} if, and only if, this array is backed by a buffer.
     */
    boolean hasBuffer();

    /**
     * Returns the component type of the backing array.
     *
     * @return component type of the backing array.
     */
    Class<? super E> getComponentType();

    /**
     * Gets the comparator that is used for sorting this array, {@code null} if array is primitive array, or if array
     * sorted according to natural ordering of elements, or {@code null} if array is not sorted.
     *
     * @return a {@code Comparator}, or {@code null} if the elements are sorted in the natural order, or {@code null} if
     * array is not sorted.
     */
    @CheckForNull Comparator<? super E> getComparator();

    /**
     * Returns the characteristics set for this array. The result is represented as ORed values from
     * {@link #NONNULL}, {@link #MUTABLE}, {@link #UNBOUNDED}, {@link #SORTED}, {@link #DISTINCT}.
     * Modification of the array contents may result in updated characteristics.
     *
     * @return the characteristics set for this array.
     */
    int characteristics();

    /**
     * Returns {@code true} if the array contain all of the given characteristics.
     *
     * <p>The Java Array Framework implementation returns true if the corresponding bits of the given characteristics are set.</p>
     *
     * @param characteristics characteristics to be tested.
     * @return {@code true} if all the specified characteristics are present, {@code false} otherwise.
     */
    default boolean hasCharacteristics(int characteristics) {
        return (characteristics() & characteristics) == characteristics;
    }
}
