/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;
import java.util.Comparator;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.function.Consumer;

import static java.util.Objects.requireNonNull;

/**
 * <p>Dynamic-size wrapper for Java Collection Framework's List.</p>
 *
 * <p>This class is used as base implementation for {@code UnboundedListArrayImpl} only.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Dynamic arrays, such as this class, always report {@code MUTABLE}.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that the array implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. Characteristics are always invalidated if
 * conditions of these characteristics are broken as result of array modification. Characteristics are only added in
 * scenarios in which their conditions can be 'tested' without traversal of elements. For example, the {@code SORTED}
 * characteristic is added if the length of an array is reduced to zero or one element, or if all elements are replaced
 * by an array that is marked as being sorted. The semantics of characteristics processing is documented for each
 * "destructive" method. Users can take full control of characteristics assignment by means of the
 * {@link MutableArray#doAction(ArrayAction)} method.</p>
 *
 * @author Leon van Zantvoort
 */
@NotThreadSafe
abstract class DynamicListArrayImpl<E> extends MutableListArrayImpl<E> implements DynamicObjectArray<E> {

    /** Counter for fail-fast error detection. */
    protected transient int modCount = 0;

    /**
     * Package visible constructor used to wrap specified {@code array}.
     *
     * @param list list to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> is used.
     * @param characteristics array characteristics.
     */
    DynamicListArrayImpl(@Nonnull List<E> list, int offset, int length, int characteristics,
                         @Nullable Comparator<? super E> comparator) {
        super(list, offset, length, characteristics, comparator);
    }

    /**
     * Fail-fast error detection. Sub classes (such as sub arrays) may override this method if they wish to implement
     * concurrent modification checking.
     *
     * @throws ConcurrentModificationException if parent array has been structurally modified.
     */
    protected void checkForComodification() {
    }

    /**
     * Updates length of this array and increases modCount.
     *
     * <p>Sub arrays must override this method and delegate the invocation this implementation by calling
     * {@code super.updateLength()}, next they must delegate the invocation to the parent array.</p>
     *
     * @param diff number of elements that are added / removed.
     */
    protected void updateLength(int diff) {
        modCount++;
        length += diff;
    }

    /**
     * This method has no use because the backing data store is a list.
     *
     * @param flag you might as well pass it Random.nextBoolean().
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void setClearReservedElements(boolean flag) {
    }

    /**
     * Returns the number of additional elements that this array can store (in the absence of memory or resource
     * constraints), or {@code Integer.MAX_VALUE} if there is no intrinsic limit.
     *
     * @return the remaining capacity.
     */
    @Override
    public int remainingCapacity() {
        return Integer.MAX_VALUE;
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     */
    @SuppressWarnings("ConstantConditions")
    @Override
    public void forEach(@Nonnull Consumer<? super E> consumer) {
        requireNonNull(consumer);
        int expectedModCount = modCount;
        Iterator<E> it = elements.listIterator(offset);
        for (int index = 0; modCount == expectedModCount && index < length; index++) {
            consumer.accept(it.next());
        }
        if (modCount != expectedModCount) {
            throw new ConcurrentModificationException();
        }
    }

    /**
     * Replaces all elements of this array by the elements of the specified array. The semantics of this method are less
     * strict than the semantics expressed by {@code MutableArray}; The length of the array that is passed as argument
     * needs to be less than or equal to this array's capacity.
     *
     * <p>The length of {@code this} array is increased or decreased by length difference between {@code this} array and
     * {@code array} argument.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the specified {@code array} reports {@code NONNULL}, {@code NONNULL} is reported after this
     *   invocation;</li>
     *   <li>if the specified {@code array} reports {@code SORTED}, {@code SORTED} is reported after this invocation if
     *   the comparator of the specified {@code array} is equal to this array's comparator;</li>
     *   <li>if the specified {@code array} reports {@code SORTED} and {@code DISTINCT}, {@code DISTINCT} is reported
     *   after this invocation if the comparator of the specified {@code array} is equal to this array's comparator.
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws NullPointerException if the specified array is null or if the element is replaced with a null value
     * and this array does not permit null elements.
     * @throws IllegalStateException if elements cannot be replaced due to capacity issues.
     */
    @Override
    public void setAll(@Nonnull Array<? extends E> array) {
        int lengthDiff = array.length() - length;
        elements.subList(offset, offset + length).clear();
        elements.addAll(offset, ArrayBridge.asList(array));
        if (lengthDiff != 0) {
            updateLength(lengthDiff);
        }
        characteristics(array.characteristics());   // All elements are replaced by new array; Copy characteristics.
        if (hasCharacteristics(SORTED)) {
            if (!Objects.equals(comparator, array.getComparator())) {
                removeCharacteristics(SORTED | DISTINCT);
            } else {
                updateCharacteristics(offset, array);
            }
        }
    }

    /**
     * Appends the specified object to the end of this array. Length of this array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param object object to to be added.
     * @throws NullPointerException if object is null and array does not permit null elements.
     * @throws IllegalStateException if element cannot be inserted due to capacity issues.
     */
    @Override
    public void add(E object) {
        int elementIndex = offset + length;
        elements.add(elementIndex, object);
        updateLength(1);
        if (object == null) {
            removeCharacteristics(NONNULL);
        }
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, object);
        }
    }

    /**
     * Appends the elements of the specified array to the end of this array. Length of {@code this} array is increased
     * by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to be added.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     * @throws IllegalStateException if element cannot be inserted due to capacity issues.
     */
    @Override
    public void addAll(@Nonnull Array<? extends E> array) {
        int arrayLength = array.length();
        if (arrayLength > 0) {
            int elementIndex = offset + length;
            elements.addAll(elementIndex, ArrayBridge.asList(array));
            updateLength(arrayLength);
            retainCharacteristics(array.characteristics());
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(elementIndex, array);
            }
        }
    }

    /**
     * Removes last element. Length of this array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @return element that is removed.
     * @throws NoSuchElementException if array is already empty.
     */
    @Override
    public E remove() {
        if (length == 0) {
            throw new NoSuchElementException();
        }
        E oldValue = elements.remove(offset + length - 1);
        updateLength(-1);
        addCharacteristics(length == 0 ? NONNULL | SORTED | DISTINCT : 0);
        return oldValue;
    }

    /**
     * Removes all elements.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>After this invocation, {@code NONNULL}, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     */
    @Override
    public void clear() {
        if (length > 0) {
            if (offset == 0 && length == elements.size()) {
                elements.clear();
            } else {
                elements.subList(offset, offset + length).clear();
            }
            updateLength(-length);
        }
        addCharacteristics(NONNULL | SORTED | DISTINCT);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        final ListIterator<E> iterator;

        /**
         * The modCount value that the iterator believes that the backing
         * array should have. If this expectation is violated, the iterator
         * has detected concurrent modification.
         */
        int expectedModCount = modCount;

        ArrayIteratorImpl(ListIterator<E> iterator) {
            this.iterator = iterator;
        }

        void checkForComodification() {
            if (modCount != expectedModCount) {
                throw new ConcurrentModificationException();
            }
        }

        @Override
        public boolean hasNext() {
            return iterator.hasNext();
        }

        @Override
        public E next() {
            checkForComodification();
            return iterator.next();
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return iterator.hasPrevious();
        }

        @Override
        public E previous() {
            checkForComodification();
            return iterator.previous();
        }

        @Override
        public int nextIndex() {
            return iterator.nextIndex();
        }

        @Override
        public int previousIndex() {
            return iterator.previousIndex();
        }

        @SuppressWarnings("unchecked")
        @Override
        public void set(E object) {
            checkForComodification();
            iterator.set(object);
            if (length > 1) {      // Unable to compare against replaced object.
                removeCharacteristics(SORTED | DISTINCT);
            }
        }

        @Override
        public void add(E object) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator());
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        rangeCheckForIterator(index);
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator(index));
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public abstract DynamicObjectArray<E> offset(int offset);
}
