/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Objects;
import java.util.RandomAccess;
import java.util.function.Consumer;

import static java.lang.System.arraycopy;
import static java.util.Objects.requireNonNull;

/**
 * <p>Immutable, fixed-size, wrapper for Java Collection Framework's List.</p>
 *
 * <p>This immutable object array is constructed by the static {@code newArray} methods of {@code Arrays}.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that this implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. The set of characteristics of immutable
 * arrays are constant.</p>
 *
 * @author Leon van Zantvoort
 */
@Immutable
class ListArrayImpl<E> extends AbstractArray<E> implements ObjectArray<E>, Serializable {

    /** backing list. */
    protected transient List<E> elements;

    /** Comparator that is used for sorting array. */
    protected Comparator<? super E> comparator;

    /**
     * Package visible constructor, to be used by static factory methods only.
     *
     * @param list backing list.
     * @param offset start index of range.
     * @param length length of range.
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> is used.
     * @param characteristics array characteristics.
     */
    ListArrayImpl(@Nonnull List<E> list, int offset, int length, int characteristics,
                  @Nullable Comparator<? super E> comparator) {
        super(offset, length, characteristics);
        requireNonNull(list);
        if (length > list.size() - offset) {
            throw new IndexOutOfBoundsException();
        }
        this.elements = list;
        this.comparator = comparator;
        if (length == 0) {
            this.characteristics |= NONNULL;
        }
    }

    /**
     * Tells whether or not this array is backed by an array.
     *
     * <p>If this method returns {@code true} then array instance may safely be obtained through {@code array()}.</p>
     *
     * @return {@code true} if, and only if, this array is backed by an array.
     */
    @Override
    public final boolean hasArray() {
        return false;
    }

    /**
     * Tells whether or not this array is backed by a buffer.
     *
     * <p>If this method returns {@code true} then buffer instance may safely be obtained through {@code buffer()}.</p>
     *
     * @return {@code true} if, and only if, this array is backed by a buffer.
     */
    @Override
    public final boolean hasBuffer() {
        return false;
    }

    /**
     * Returns the component type of the backing array.
     *
     * @return component type of the backing array.
     */
    @Override
    public final Class<? super E> getComponentType() {
        return Object.class;
    }

    /**
     * Gets the comparator that is used for sorting this array, {@code null} if array is primitive array, or if array
     * sorted according to natural ordering of elements, or {@code null} if array is not sorted.
     *
     * @return a {@code Comparator}, or {@code null} if the elements are sorted in the natural order, or {@code null} if
     * array is not sorted.
     */
    @Override
    public final @CheckForNull Comparator<? super E> getComparator() {
        return comparator;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    protected int filterCharacteristics(int characteristics) {
        return characteristics & ~(MUTABLE | BOUNDED | UNBOUNDED);
    }

    /**
     * Returns the object found at {@code index}.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @SuppressWarnings("unchecked")
    @Override
    public @Nullable E get(int index) {
        rangeCheck(index);
        return elements.get(offset + index);
    }

    /**
     * Returns the object found at {@code index}. This method allows for fast
     * failure detection.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if null reference is found at index.
     */
    @SuppressWarnings("unchecked")
    @Override
    public @Nonnull E getNonnull(int index) {
        rangeCheck(index);
        E value = elements.get(offset + index);
        requireNonNull(value, "Assertion error, nonnull reference expected.");
        return value;
    }

    /**
     * Returns {@code true} if this array contains the specified object. This method supports boxed primitives. This
     * means that if this array represents a primitive array, this method returns {@code true} if this array contains
     * the unboxed version of specified object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object object whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified object.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public boolean contains(Object object) {
        if (hasCharacteristics(SORTED) && elements instanceof RandomAccess) {
            // Optimization only possible if list supports fast random access.
            Comparator<? super E> comparator = getComparator();
             // No NullPointerException in case of natural ordering.
            return !(object == null && comparator == null) &&
                    Arrays.binarySearch(elements, offset, length, (E) object, comparator) >= 0;
        }
        final int toIndex = offset + length;
        final ListIterator<E> it = elements.listIterator(offset);
        for (int index = offset; index < toIndex; index++) {
            if (Objects.equals(it.next(), object)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns {@code true} if this array contains all of the elements in the specified {@code array}. Primitives and
     * their object counterparts are regarded equal in this respect.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array is sorted, otherwise all elements are
     * traversed until the objects in specified array are found.</p>
     *
     * @param array array to be checked for containment in this array.
     * @return {@code true} if this array contains all of the elements in the specified array.
     * @throws NullPointerException if the specified array is null, or if the specified array contains null elements and
     * array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with elements.
     */
    @Override
    public boolean containsAll(@Nonnull Array<?> array) {
        for (Object o : array) {
            if (!contains(o)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns the index of the first occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public int indexOf(Object object) {
        if (hasCharacteristics(SORTED) && elements instanceof RandomAccess) {
            // Optimization only possible if list supports fast random access.
            Comparator<? super E> comparator = getComparator();
            if (object == null && comparator == null) {  // No NullPointerException in case of natural ordering.
                return -1;
            }
            int index = Arrays.binarySearch(elements, offset, length, (E) object, comparator);
            if (index >= 0) {
                return Arrays.binarySearchFirst(elements, offset, length, index, comparator);
            }
            return -1;
        }
        final int toIndex = offset + length;
        final ListIterator<E> it = elements.listIterator(offset);
        for (int index = offset; index < toIndex; index++) {
            if (Objects.equals(it.next(), object)) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Returns the index of the last occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified object is found.</p>
     *
     * @param object object whose highest index in this array is to be returned.
     * @return highest index of object in array or {@code -1} if value is not found.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public int lastIndexOf(Object object) {
        if (hasCharacteristics(SORTED) && elements instanceof RandomAccess) {
            // Optimization only possible if list supports fast random access.
            Comparator<? super E> comparator = getComparator();
            if (object == null && comparator == null) {  // No NullPointerException in case of natural ordering.
                return -1;
            }
            int index = Arrays.binarySearch(elements, offset, length, (E) object, comparator);
            if (index >= 0) {
                return Arrays.binarySearchLast(elements, offset, length, index, comparator);
            }
            return -1;
        }
        final int toIndex = offset + length;
        final ListIterator<E> it = elements.listIterator(toIndex);
        for (int index = toIndex - 1; index >= offset; index--) {
            if (Objects.equals(it.previous(), object)) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    public Object[] toArray() {
        return elements.subList(offset, offset + length).toArray();
    }

    /**
     * Returns an array containing all of the reference of {@code this} array. If the references fit in the
     * specified array, it is returned therein. Otherwise, a new array of the same object type is allocated.
     *
     * @param <T> custom element type.
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("unchecked")
    @Override
    public <T> T[] toArray(@Nonnull T[] array) {
        requireNonNull(array);
        //noinspection SuspiciousToArrayCall
        return elements.subList(offset, offset + length).toArray(array);
    }

    /**
     * Stores elements from this array into the specified array.
     *
     * @param <T> custom element type.
     * @param array the array into which the elements of this array are to be stored.
     * @param offset the offset within the specified array of the first element to be written.
     * @param length the number of elements to be stored. Must be smaller than length of this array.
     * @throws IndexOutOfBoundsException if offset and length parameters are invalid.
     */
    @Override
    public <T> void toArray(@Nonnull T[] array, int offset, int length) {
        Object[] a = elements.subList(this.offset, this.offset + length).toArray();
        arraycopy(a, 0, array, offset, length);
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     */
    @Override
    public void forEach(@Nonnull Consumer<? super E> consumer) {
        requireNonNull(consumer);
        Iterator<E> it = elements.listIterator(offset);
        for (int index = 0; index < length; index++) {
            consumer.accept(it.next());
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public ObjectArray<E> offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new ListArrayImpl<>(elements, newOffset, newLength, characteristics, comparator);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public ObjectArray<E> length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new ListArrayImpl<>(elements, offset, length, characteristics, comparator);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    @SuppressWarnings("unchecked")
    public ObjectArray<E> subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new ListArrayImpl<>(elements, newOffset, newLength, characteristics, comparator);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        final ListIterator<E> iterator;

        ArrayIteratorImpl(ListIterator<E> iterator) {
            this.iterator = iterator;
        }

        @Override
        public boolean hasNext() {
            return iterator.hasNext();
        }

        @Override
        public E next() {
            return iterator.next();
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return iterator.hasPrevious();
        }

        @Override
        public E previous() {
            return iterator.previous();
        }

        @Override
        public int nextIndex() {
            return iterator.nextIndex();
        }

        @Override
        public int previousIndex() {
            return iterator.previousIndex();
        }

        @Override
        public void set(E reference) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void add(E reference) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator());
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator(index));
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public String toString() {
        int iMax = offset + length - 1;
        if (iMax - offset == -1) {
            return "[]";
        }
        StringBuilder b = new StringBuilder();
        b.append('[');
        for (Iterator<E> it = iterator(); it.hasNext();) {
            b.append(it.next());
            if (!it.hasNext()) {
                return b.append(']').toString();
            }
            b.append(", ");
        }
        return b.toString();
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public int hashCode() {
        int result = 1;
        //noinspection ForLoopReplaceableByForEach
        for (Iterator<E> it = iterator(); it.hasNext();) {
            Object element = it.next();
            result = 31 * result + (element == null ? 0 : element.hashCode());
        }
        return result;
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj instanceof Array) {
            Array<?> array = (Array<?>) obj;
            if (length != array.length()) {
                return false;
            }
            // Fastest: direct array access.
            Object[] backingArray = NativeArrays.array(array);
            if (backingArray != null) {
                final ArrayIterator<?> iterator = iterator();
                for (int index = array.offset(), toIndex = index + length; index < toIndex; index++) {
                    if (!Objects.equals(iterator.next(), backingArray[index])) {
                        return false;
                    }
                }
                return true;
            }
            // Alternative.
            return Arrays.equals(this, array);
        }
        return false;
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    Object writeReplace() {
        return new SerializationProxy<>(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy<E> implements Serializable {
        static final long serialVersionUID = -6316889894958073614L;

        transient int offset;
        final int length;
        final int characteristics;
        final Comparator<? super E> comparator;
        transient List<E> elements;

        SerializationProxy(ListArrayImpl<E> array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.comparator = array.comparator;
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            if (offset == 0 && length == elements.size()) {
                s.writeObject(elements);
            } else {
                s.writeObject(elements.subList(offset, offset + length));
            }
        }

        @SuppressWarnings("unchecked")
        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = (List<E>) s.readObject();
        }

        @SuppressWarnings("unchecked")
        private Object readResolve() {
            return new ListArrayImpl<>(elements, 0, length, characteristics, comparator);
        }
    }
}