/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

/**
 * <p>Immutable array of object references.</p>
 *
 * <p>Object arrays are constructed by the static interface methods of this interface. The {@code copyOf} method returns
 * a new array representing a shallow copy of the array argument. The {@code unsafeValueOf}
 * methods wrap a new {@code ObjectArray} around a native array. It is imperative that this native array is not modified
 * outside the scope of {@code ObjectArray}, as this would break the immutable contract. The
 * {@code unsafeValueOf} method is used to prevent unnecessary array copies. It is the caller's responsibility not to 
 * violate the contract.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * The set of characteristics of immutable arrays are constant.</p>
 *
 * @param <E> element type.
 * @author Leon van Zantvoort
 */
@Immutable
public interface ObjectArray<E> extends Array<E> {

    /** Empty array. */
    @SuppressWarnings("rawtypes")
    public static final ObjectArray EMPTY = unsafeValueOf();

    /**
     * Returns an empty array for the specified object type.
     *
     * @param <E> target type.
     * @return an empty array.
     */
    @SuppressWarnings("unchecked")
    static <E> ObjectArray<E> emptyArray() {
        return (ObjectArray<E>) EMPTY;
    }

    /**
     * Creates an immutable copy of the specified {@code array}, backed by a native array. If the specified
     * {@code array} is mutable, changes to the mutable array are not reflected by this immutable copy.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements.</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code MUTABLE} and {@code BOUNDED}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param array array to be copied.
     * @return immutable array of objects.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("unchecked")
    static <E> ObjectArray<E> copyOf(@Nonnull Array<E> array) {
        E[] backingArray = (E[]) NativeArrays.array(array);
        if (backingArray != null) {
            return Arrays.newArray(backingArray, array.offset(), array.length(), array.characteristics(),
                    array.getComparator(), array.hasCharacteristics(MUTABLE));
        }
        return Arrays.newArray((E[]) NativeArrays.toArray(array), 0, array.length(), array.characteristics(),
                array.getComparator(), false);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be copied.
     * @param <E> element type.
     * @return immutable array.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    @SafeVarargs
    static <E> ObjectArray<E> copyOf(@Nonnull E... array) {
        return copyOf(array, 0, array.length);
    }

    /**
     * <p>Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.</p>
     *
     * <p>Note that the backing array will have the same length as {@code array}.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL} if the specified {@code length} is zero;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param <E> element type.
     * @return immutable array.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    static <E> ObjectArray<E> copyOf(@Nonnull E[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, null, true);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array, it must therefore not be modified after this method returns as this would violate
     * the resulting array's immutability and characteristics contract.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param <E> element type.
     * @return immutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    @SafeVarargs
    static <E> ObjectArray<E> unsafeValueOf(@Nonnull E... array) {
        return unsafeValueOf(array, 0, array.length);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array, it must therefore not be modified after this method returns as this would violate
     * the resulting array's immutability and characteristics contract.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param <E> element type.
     * @return immutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    static <E> ObjectArray<E> unsafeValueOf(@Nonnull E[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, null, false);
    }

    /**
     * Returns the object found at {@code index}.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    @Nullable E get(int index);

    /**
     * Returns the object found at {@code index}. This method allows for fast
     * failure detection.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if null reference is found at index.
     */
    @Nonnull E getNonnull(int index);

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    Object[] toArray();

    /**
     * Returns an array containing all of the reference of {@code this} array. If the references fit in the
     * specified array, it is returned therein. Otherwise, a new array of the same object type is allocated.
     *
     * @param <T> custom element type.
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    <T> T[] toArray(@Nonnull T[] array);

    /**
     * Stores elements from this array into the specified array.
     *
     * @param <T> custom element type.
     * @param array the array into which the elements of this array are to be stored.
     * @param offset the offset within the specified array of the first element to be written.
     * @param length the number of elements to be stored. Must be smaller than length of this array.
     * @throws IndexOutOfBoundsException if offset and length parameters are invalid.
     */
    <T> void toArray(@Nonnull T[] array, int offset, int length);

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    ObjectArray<E> offset(int offset);  // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    ObjectArray<E> length(int length);  // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    ObjectArray<E> subArray(int fromIndex, int toIndex);    // More specific return type.
}