/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import org.squarebrackets.function.BooleanConsumer;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.NotThreadSafe;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;

import static java.lang.System.arraycopy;

/**
 * <p>Capacity unbounded array of boolean primitives.</p>
 *
 * <p>This unbounded boolean array is constructed by the static interface methods of {@code UnboundedBooleanArray}. The
 * {@code newInstance} method returns a new empty array. The {@code copyOf} method returns a new array representing a
 * deep copy of the array argument. The {@code valueOf} methods wrap a new {@code UnboundedBooleanArray} around a native
 * array.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Unbounded primitive arrays, such as this class, always report {@code UNBOUNDED}, {@code MUTABLE} and
 * {@code NONNULL}.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that this implementation determines which set of characteristics apply to its elements without traversing these
 * elements, in other words, without incurring a performance penalty. Characteristics are always invalidated if
 * conditions of these characteristics are broken as result of array modification. Characteristics are only added in
 * scenarios in which their conditions can be 'tested' without traversal of elements. For example, the {@code SORTED}
 * characteristic is added if the length of an array is reduced to zero or one element, or if all elements are replaced
 * by an array that is marked as being sorted. The semantics of characteristics processing is documented for each
 * "destructive" method. Users can take full control of characteristics assignment by means of the
 * {@link MutableArray#doAction(ArrayAction)} method.</p>
 *
 * @author Leon van Zantvoort
 */
@NotThreadSafe
class UnboundedBooleanArrayImpl extends DynamicBooleanArrayImpl implements UnboundedBooleanArray {

    /**
     * Package visible constructor used to wrap specified {@code array}.
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param characteristics array characteristics.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    UnboundedBooleanArrayImpl(@Nonnull boolean[] array, int offset, int length, int characteristics, boolean copy) {
        super(array, offset, length, characteristics, copy);
    }

    /**
     * Replaces the underling array with the specified {@code array}.
     *
     * <p>Sub arrays must override this method and delegate the invocation this implementation by calling
     * {@code super.setArray()}, next they must delegate the invocation to the parent array.</p>
     *
     * @param array new backing array.
     */
    protected void setArray(boolean[] array) {
        elements = array;
    }

    /**
     * Returns the length of the root array. If {@code this} array is the root array, this method returns the same
     * value is {@code length()}.
     *
     * <p>Sub arrays must override this method and delegate the invocation to the parent.</p>
     *
     * @return length of the root array.
     */
    protected int rootLength() {
        return length;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    @Override
    protected int filterCharacteristics(int characteristics) {
        return (characteristics | NONNULL | MUTABLE | UNBOUNDED) & ~BOUNDED;
    }

    /**
     * Ensure that backing array can accommodate at least {@code capacity} number of elements.
     *
     * @param capacity minimal capacity.
     */
    private void ensureCapacity(int capacity) {
        int reservedCapacity = elements.length;
        if (capacity > reservedCapacity) {
            int newCapacity = reservedCapacity + (reservedCapacity >> 1);
            if (newCapacity - capacity < 0) {
                newCapacity = capacity;
            }
            if (newCapacity - Arrays.MAX_ARRAY_SIZE > 0) {
                newCapacity = Arrays.hugeCapacity(capacity);
            }
            setArray(java.util.Arrays.copyOf(elements, newCapacity));
        }
    }

    /**
     * Returns the number of additional elements that this array can store (in the absence of memory or resource
     * constraints), or {@code Integer.MAX_VALUE} if there is no intrinsic limit.
     *
     * @return the remaining capacity.
     */
    @Override
    public int remainingCapacity() {
        return Integer.MAX_VALUE;
    }

    /**
     * Appends the specified boolean value to the end of this array. Length of this array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param value boolean value to to be added.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void addBoolean(boolean value) {
        ensureCapacity(rootLength() + 1);
        int elementIndex = offset + length;
        elements[elementIndex] = value;
        updateLength(1);
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, value);
        }
    }

    /**
     * Inserts the specified object at the specified {@code index}, moving elements at and after {@code index} to the
     * right. Length of {@code this} array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param object object to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if specified {@code object} is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void add(int index, Boolean object) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        addBoolean(index, object);
    }

    /**
     * Inserts the specified boolean value at the specified {@code index}, moving elements at and after {@code index} to the
     * right. Length of {@code this} array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param value value to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    public void addBoolean(int index, boolean value) {
        rangeCheckForAdd(index);
        int rootLength = rootLength();
        int elementIndex = offset + index;
        ensureCapacity(rootLength + 1);
        if (rootLength != elementIndex) {    // Item added in the middle.
            int len;
            if ((len = rootLength - elementIndex) > 0) {
                arraycopy(elements, elementIndex, elements, elementIndex + 1, len);
            }
        }
        elements[elementIndex] = value;
        updateLength(1);
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, value);
        }
    }

    /**
     * Appends the elements of the specified array to the end of this array. Length of {@code this} array is increased
     * by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to be added.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void addAll(@Nonnull Array<? extends Boolean> array) {
        int arrayLength = array.length();
        if (arrayLength > 0) {
            ensureCapacity(rootLength() + arrayLength);
            int elementIndex = offset + length;
            BooleanArray cast = ArrayCast.toBooleanArray(array);
            if (cast != null) {
                cast.toArray(elements, elementIndex, arrayLength);
            } else {
                arraycopy(NativeArrays.toBooleanArray(array), 0, elements, elementIndex, arrayLength);
            }
            updateLength(arrayLength);
            retainCharacteristics(array.characteristics());
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(elementIndex, array);
            }
        }
    }

    /**
     * Inserts the elements of the specified array at the specified {@code index}, moving elements at and after
     * {@code index} to the right. Length of {@code this} array is increased by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param array array to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void addAll(int index, @Nonnull Array<? extends Boolean> array) {
        rangeCheckForAdd(index);
        int arrayLength = array.length();
        if (arrayLength > 0) {
            int rootLength = rootLength();
            ensureCapacity(rootLength + arrayLength);
            int elementIndex = offset + index;
            if (rootLength != elementIndex) {    // Items added in the middle.
                int len;
                if ((len = rootLength - elementIndex) > 0) {
                    arraycopy(elements, elementIndex, elements, elementIndex + arrayLength, len);
                }
            }
            BooleanArray cast = ArrayCast.toBooleanArray(array);
            if (cast != null) {
                cast.toArray(elements, elementIndex, arrayLength);
            } else {
                arraycopy(NativeArrays.toBooleanArray(array), 0, elements, elementIndex, arrayLength);
            }
            updateLength(arrayLength);
            retainCharacteristics(array.characteristics());
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(elementIndex, array);
            }
        }
    }

    /**
     * Replaces all elements of this array by the elements of the specified array. The semantics of this method are
     * similar to the semantics expressed by {@code DynamicArray}; Since the capacity of this array is unbounded, the
     * array that is passed can be of any length.
     *
     * <p>The length of {@code this} array is increased or decreased by length difference between {@code this} array and
     * {@code array} argument.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the specified {@code array} reports {@code SORTED}, {@code SORTED} is reported after this invocation if
     *   the comparator of the specified {@code array} is equal to this array's comparator;</li>
     *   <li>if the specified {@code array} reports {@code SORTED} and {@code DISTINCT}, {@code DISTINCT} is reported
     *   after this invocation if the comparator of the specified {@code array} is equal to this array's comparator.
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws NullPointerException if the specified array is null or if the element is replaced with a null value.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void setAll(@Nonnull Array<? extends Boolean> array) {
        int arrayLength = array.length();
        int lengthDiff = arrayLength - length;
        if (lengthDiff != 0) {
            int rootLength = rootLength();
            if (lengthDiff > 0) {
                ensureCapacity(rootLength + lengthDiff);
            }
            int len;
            if ((len = rootLength - (offset + length)) > 0) {
                arraycopy(elements, offset + length, elements, offset + arrayLength, len);
            }
        }
        if (arrayLength > 0) {
            BooleanArray cast = ArrayCast.toBooleanArray(array);
            if (cast != null) {
                cast.toArray(elements, offset, arrayLength);
            } else {
                arraycopy(NativeArrays.toBooleanArray(array), 0, elements, offset, arrayLength);
            }
        }
        if (lengthDiff != 0) {
            updateLength(lengthDiff);
        }
        characteristics(array.characteristics());   // All elements are replaced by new array; Copy characteristics.
        if (hasCharacteristics(SORTED)) {
            if (array.getComparator() != null) {
                removeCharacteristics(SORTED | DISTINCT);
            } else {
                updateCharacteristics(offset, array);
            }
        }
    }

    /**
     * Removes last element. Length of this array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @return element that is removed.
     * @throws NoSuchElementException if array is already empty.
     */
    @Override
    public boolean booleanRemove() {
        if (length == 0) {
            throw new NoSuchElementException();
        }
        boolean oldValue = elements[offset + length - 1];
        shiftElements(length - 1, 1);
        return oldValue;
    }

    /**
     * Removes last element. Length of this array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param index removal index, relative to the offset of the array.
     * @return element that is removed.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public Boolean remove(int index) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        return booleanRemove(index);
    }

    /**
     * Removes element at specified {@code index}, moving all elements after {@code index} to the left. Length of
     * {@code this} array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param index removal index, relative to the offset of the array.
     * @return element that is removed.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    public boolean booleanRemove(int index) {
        rangeCheck(index);
        boolean oldValue = elements[offset + index];
        shiftElements(index, 1);
        return oldValue;
    }

    /**
     * Removes the first occurrence of the specified object from this array if it is present. If this array does not
     * contain the object, it is unchanged. This method supports boxed primitives. This means that if this array
     * represents a primitive array, this method removes the first primitive element that is the unboxed counterpart of
     * the object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics to lookup the element to be removed, otherwise all elements are traversed until the specified
     * object is found.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param object object to be removed.
     * @return {@code true} if this array contained the specified object.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public boolean remove(Object object) {
        if (object instanceof Boolean) {
            if (Tripwire.ENABLED) {
                Tripwire.trip(getClass());
            }
            return removeBoolean((Boolean) object);
        }
        return false;
    }

    /**
     * Removes the first occurrence of the specified value from this array if it is present. If this array does not
     * contain the value, it is unchanged.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics to lookup the element to be removed, otherwise all elements are traversed until the specified
     * value is found.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param value value to be removed from this array.
     * @return {@code true} if this array contained the specified value.
     */
    @Override
    public boolean removeBoolean(boolean value) {
        if (hasCharacteristics(SORTED)) {
            int index = Arrays.binarySearch(this, value);
            if (index >= 0) {
                booleanRemove(index);
                return true;
            }
            return false;
        }
        for (int index = offset, toIndex = index + length; index < toIndex; index++) {
            if (elements[index] == value) {
                booleanRemove(index - offset);
                return true;
            }
        }
        return false;
    }

    /**
     * Removes all of this array's elements that are also contained in the specified array. Primitives and their object
     * counterparts are regarded equal in this respect. After this call returns, this array will contain no elements
     * in common with the specified array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param array array containing elements to be removed from this array.
     * @return {@code true} if this array changed as a result of the call.
     * @throws NullPointerException if the specified array is null.
     * @throws ClassCastException if specified array is sorted and specified array's comparator not compatible with
     * with this array's elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public boolean removeAll(@Nonnull Array<?> array) {
        BooleanArray cast = ArrayCast.toBooleanArray(array);
        if (cast != null) {
            return removeBoolean(cast, false);
        }
        return remove(array, false);
    }

    /**
     * Retains only the elements in this array that are contained in the specified array. In other words, removes from
     * this array all of its elements that are not contained in the specified array. Primitives and their object
     * counterparts are regarded equal in this respect.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param array array containing elements to be retained in this array.
     * @return {@code true} if this array changed as a result of the call.
     * @throws NullPointerException if the specified array is null.
     * @throws ClassCastException if specified array is sorted and specified array's comparator not compatible with
     * with this array's elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public boolean retainAll(@Nonnull Array<?> array) {
        BooleanArray cast = ArrayCast.toBooleanArray(array);
        if (cast != null) {
            return removeBoolean(cast, true);
        }
        return remove(array, true);
    }

    /**
     * Removes or retains elements from this array.
     *
     * @param keys elements to be removed or retained.
     * @param complement {@code true} for retain, {@code false} for remove.
     * @return {@code true} if elements are removed as result of this call, {@code false} otherwise.
     */
    private boolean remove(Array<?> keys, boolean complement) {
        if (!complement && hasCharacteristics(DISTINCT) && length > keys.length()) {    // Optimized scenario.
            boolean modified = false;
            for (Object key : keys) {
                modified |= remove(key);
            }
            return modified;
        }
        int rIndex = offset, wIndex = rIndex;
        int toIndex = offset + length;
        boolean[] a = elements;
        for (; rIndex < toIndex; rIndex++) {
            if (keys.contains(a[rIndex]) == complement) {
                a[wIndex++] = a[rIndex];
            }
        }
        int removed = toIndex - wIndex;
        if (removed > 0) {
            shiftElements(length - 1, removed);
            return true;
        }
        return false;
    }

    /**
     * Removes or retains elements from this array.
     *
     * @param keys elements to be removed or retained.
     * @param complement {@code true} for retain, {@code false} for remove.
     * @return {@code true} if elements are removed as result of this call, {@code false} otherwise.
     */
    private boolean removeBoolean(BooleanArray keys, boolean complement) {
        if (!complement && hasCharacteristics(DISTINCT) && length > keys.length()) {    // Optimized scenario.
            boolean modified = false;
            for (ArrayIterator.OfBoolean it = keys.iterator(); it.hasNext();) {
                modified |= removeBoolean(it.nextBoolean());
            }
            return modified;
        }
        int rIndex = offset, wIndex = rIndex;
        int toIndex = offset + length;
        boolean[] a = elements;
        for (; rIndex < toIndex; rIndex++) {
            if (keys.containsBoolean(a[rIndex]) == complement) {
                a[wIndex++] = a[rIndex];
            }
        }
        int removed = toIndex - wIndex;
        if (removed > 0) {
            shiftElements(length - 1, removed);
            return true;
        }
        return false;
    }

    /**
     * Removes all elements.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>After this invocation {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void clear() {
        shiftElements(Math.max(0, length - 1), length);
    }

    /**
     * Replaces backing array by a new array that has the same length as the root array, if and only if the
     * backing is greater than the root array.
     */
    @Override
    public void compact() {
        int rootLength = rootLength();
        if (rootLength != elements.length) {
            setArray(java.util.Arrays.copyOf(elements, rootLength));
        }
    }

    /**
     * Removes all duplicates from this array. The order of elements may be altered as result of this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, adhere to the characteristics contract:</p>
     *
     * <ul>
     *   <li>{@code SORTED} and {@code DISTINCT} are reported after this invocation.</li>
     * </ul>
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void distinct() {
        sort();
        if (!hasCharacteristics(DISTINCT)) {
            if (length > 1) {
                int rIndex = offset + 1, wIndex = rIndex;
                int toIndex = offset + length;
                boolean previous = elements[offset];
                while (rIndex < toIndex) {
                    boolean next = elements[rIndex++];
                    if (next != previous) {
                        elements[wIndex++] = next;
                        previous = next;
                    }
                }
                int removed = rIndex - wIndex;
                if (removed > 0) {
                    shiftElements(length - 1, removed);
                }
            }
            addCharacteristics(DISTINCT);
        }
    }

    /**
     * Shifts elements after removal, decreases array length and updates characteristics.
     *
     * @param index removal index, relative to the offset of the array.
     * @param removed number of elements that need to be shifted.
     */
    private void shiftElements(int index, int removed) {
        assert removed >= 0;
        if (removed > 0) {
            int len;
            int elementIndex = offset + index + 1;
            if ((len = rootLength() - elementIndex) > 0) {
                arraycopy(elements, elementIndex, elements, elementIndex - removed, len);
            }
            updateLength(-removed);
        }
        if (length <= 1) {
            addCharacteristics(SORTED | DISTINCT);
        }
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SubUnboundedBooleanArrayImpl extends UnboundedBooleanArrayImpl {

        final UnboundedBooleanArrayImpl parent;

        SubUnboundedBooleanArrayImpl(@Nonnull UnboundedBooleanArrayImpl parent, int offset, int length) {
            super(parent.elements, offset, length, parent.characteristics, false);
            this.parent = parent;
            this.modCount = parent.modCount;
        }

        @Override
        public void setClearReservedElements(boolean flag) {
            parent.setClearReservedElements(flag);  // super.setClearReservedElements() must not be called.
        }

        @Override
        protected boolean clearReservedElements() {
            return parent.clearReservedElements();
        }

        @Override
        protected void checkForComodification() {
            if (modCount != parent.modCount) {
                throw new ConcurrentModificationException();
            }
            parent.checkForComodification();
        }

        @Override
        protected ArrayContext<Boolean> getArrayContext() {
            return parent.inAction ? null : super.getArrayContext();
        }

        @Override
        protected void setArray(boolean[] array) {
            super.setArray(array);
            parent.setArray(array);
        }

        @Override
        protected int rootLength() {
            return parent.rootLength();
        }

        @Override
        protected void updateLength(int diff) {
            super.updateLength(diff);
            parent.updateLength(diff);
        }

        @Override
        protected void updateCharacteristics(int elementIndex, boolean value) {
            super.updateCharacteristics(elementIndex, value);
            if (parent.hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, value);
            }
        }

        @Override
        protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends Boolean> array) {
            super.updateCharacteristics(elementIndex, array);
            if (parent.hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, array);
            }
        }

        @Override
        public int characteristics(int characteristics) {
            // Parent array retains characteristics that are shared with sub array. Additional characteristics
            // assigned to sub array are not set for parent array, as it is not guaranteed that these
            // characteristics hold for parent array.
            parent.retainCharacteristics(characteristics);
            return super.characteristics(characteristics);
        }

        @Override
        public ArrayIterator.OfBoolean iterator() {
            checkForComodification();
            return super.iterator();
        }

        @Override
        public ArrayIterator.OfBoolean iterator(int index) {
            checkForComodification();
            return super.iterator(index);
        }

        @Override
        public void forEachBoolean(@Nonnull BooleanConsumer consumer) {
            checkForComodification();
            super.forEachBoolean(consumer);
        }

        @Override
        public boolean getBoolean(int index) {
            checkForComodification();
            return super.getBoolean(index);
        }

        @Override
        public boolean setBoolean(int index, boolean value) {
            checkForComodification();
            return super.setBoolean(index, value);
        }

        @Override
        public void setAll(@Nonnull Array<? extends Boolean> array) {
            checkForComodification();
            super.setAll(array);
        }

        @Override
        public void addBoolean(boolean value) {
            checkForComodification();
            super.addBoolean(length, value);   // Element might be added in the middle, use index enabled method instead.
        }

        @Override
        public void addBoolean(int index, boolean value) {
            checkForComodification();
            super.addBoolean(index, value);
        }

        @Override
        public void addAll(@Nonnull Array<? extends Boolean> array) {
            checkForComodification();
            super.addAll(length, array);    // Elements might be added in the middle, use index enabled method instead.
        }

        @Override
        public void addAll(int index, @Nonnull Array<? extends Boolean> array) {
            checkForComodification();
            super.addAll(index, array);
        }

        @Override
        public boolean booleanRemove() {
            checkForComodification();
            return super.booleanRemove();
        }

        @Override
        public boolean removeBoolean(boolean value) {
            checkForComodification();
            return super.removeBoolean(value);
        }

        @Override
        public boolean booleanRemove(int index) {
            checkForComodification();
            return super.booleanRemove(index);
        }

        @Override
        public boolean removeAll(@Nonnull Array<?> array) {
            checkForComodification();
            return super.removeAll(array);
        }

        @Override
        public boolean retainAll(@Nonnull Array<?> array) {
            checkForComodification();
            return super.retainAll(array);
        }

        @Override
        public void clear() {
            checkForComodification();
            super.clear();
        }

        @Override
        public boolean[] toArray() {
            checkForComodification();
            return super.toArray();
        }

        @Override
        public boolean[] toArray(@Nonnull boolean[] array) {
            checkForComodification();
            return super.toArray(array);
        }

        @Override
        public void toArray(@Nonnull boolean[] array, int offset, int length) {
            checkForComodification();
            super.toArray(array, offset, length);
        }

        // All default implementations (defender methods) will be checked for concurrent modification either through
        // methods listed above, or through backing array access below.

        @Override
        public boolean[] array() {
            checkForComodification();
            return super.array();
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public UnboundedBooleanArray offset(int offset) {
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new SubUnboundedBooleanArrayImpl(this, newOffset, newLength);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public UnboundedBooleanArray length(int length) {
        subArrayCheck(offset, length);
        return new SubUnboundedBooleanArrayImpl(this, offset, length);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public UnboundedBooleanArray subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        subArrayCheck(newOffset, newLength);
        return new SubUnboundedBooleanArrayImpl(this, newOffset, newLength);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator.OfBoolean {
        int index;
        int lastIndex = -1;

        /**
         * The modCount value that the iterator believes that the backing
         * array should have. If this expectation is violated, the iterator
         * has detected concurrent modification.
         */
        int expectedModCount = modCount;

        ArrayIteratorImpl() {
            this(0);
        }

        ArrayIteratorImpl(int index) {
            this.index = offset + index;
        }

        void checkForComodification() {
            if (modCount != expectedModCount) {
                throw new ConcurrentModificationException();
            }
        }

        @Override
        public boolean hasNext() {
            return index < offset + length;
        }

        @Override
        public boolean nextBoolean() {
            checkForComodification();
            int i = index;
            if (i >= offset + length) {
                throw new NoSuchElementException();
            }
            boolean[] a = elements;
            if (i >= a.length) {
                throw new ConcurrentModificationException();
            }
            index = i + 1;
            return a[lastIndex = i];
        }

        @Override
        public void remove() {
            if (lastIndex - offset < 0) {
                throw new IllegalStateException();
            }
            checkForComodification();
            try {
                UnboundedBooleanArrayImpl.this.booleanRemove(lastIndex - offset);
                index = lastIndex;
                lastIndex = -1;
                expectedModCount = modCount;
            } catch (IndexOutOfBoundsException e) {
                throw new ConcurrentModificationException();
            }
        }

        @Override
        public boolean hasPrevious() {
            return index > offset;
        }

        @Override
        public boolean previousBoolean() {
            checkForComodification();
            int i = index - 1;
            if (i < offset) {
                throw new NoSuchElementException();
            }
            boolean[] a = elements;
            if (i >= a.length) {
                throw new ConcurrentModificationException();
            }
            index = i;
            return a[lastIndex = i];
        }

        @Override
        public int nextIndex() {
            return index;
        }

        @Override
        public int previousIndex() {
            return index - 1;
        }

        @Override
        public void setBoolean(boolean value) {
            if (lastIndex - offset < 0) {
                throw new IllegalStateException();
            }
            checkForComodification();
            try {
                UnboundedBooleanArrayImpl.this.set(lastIndex - offset, value);
            } catch (IndexOutOfBoundsException e) {
                throw new ConcurrentModificationException();
            }
        }

        @Override
        public void addBoolean(boolean value) {
            checkForComodification();
            try {
                int i = index;
                UnboundedBooleanArrayImpl.this.add(i - offset, value);
                index = i + 1;
                lastIndex = -1;
                expectedModCount = modCount;
            } catch (IndexOutOfBoundsException e) {
                throw new ConcurrentModificationException();
            }
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator.OfBoolean iterator() {
        return new ArrayIteratorImpl();
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator.OfBoolean iterator(int index) {
        rangeCheckForIterator(index);
        return new ArrayIteratorImpl(index);
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    @Override
    Object writeReplace() {
        return new SerializationProxy(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy implements Serializable {
        static final long serialVersionUID = 5789991360196126103L;

        transient int offset;
        final int length;
        final int characteristics;
        final boolean clearReservedElements;
        transient boolean[] elements;

        SerializationProxy(UnboundedBooleanArrayImpl array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.clearReservedElements = array.clearReservedElements();
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            for (int index = offset, toIndex = index + length; index < toIndex; index++) {
                s.writeBoolean(elements[index]);
            }
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = new boolean[length];
            for (int index = 0; index < length; index++) {
                elements[index] = s.readBoolean();
            }
        }

        private Object readResolve() {
            UnboundedBooleanArrayImpl array = new UnboundedBooleanArrayImpl(elements, 0, length, characteristics, false);
            array.setClearReservedElements(clearReservedElements);
            return array;
        }
    }
}
