/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets.function;

import static java.util.Objects.requireNonNull;

/**
 * Determines if the {@code boolean} input value matches some criteria.
 */
@FunctionalInterface
public interface BooleanPredicate {

    /**
     * Returns {@code true} if the input value matches some criteria.
     *
     * @param value the value to be tested
     * @return {@code true} if the input value matches some criteria, otherwise
     * {@code false}
     */
    boolean test(boolean value);

    /**
     * Returns a predicate which evaluates to {@code true} only if this
     * predicate and the provided predicate both evaluate to {@code true}. If
     * this predicate returns {@code false} then the remaining predicate is not
     * evaluated.
     *
     * <p>Any exceptions thrown by either {@code test} method are relayed
     * to the caller; if performing first operation throws an exception, the
     * second operation will not be performed.
     *
     * @param p a predicate which will be logically-ANDed with this predicate
     * @return a new predicate which returns {@code true} only if both
     * predicates return {@code true}
     * @throws NullPointerException if p is null
     */
    default BooleanPredicate and(BooleanPredicate p) {
        requireNonNull(p);
        return (value) -> test(value) && p.test(value);
    }

    /**
     * Returns a predicate which negates the result of this predicate.
     *
     * @return a new predicate who's result is always the opposite of this
     * predicate
     */
    default BooleanPredicate negate() {
        return (value) -> !test(value);
    }

    /**
     * Returns a predicate which evaluates to {@code true} if either this
     * predicate or the provided predicate evaluates to {@code true}. If this
     * predicate returns {@code true} then the remaining predicate is not
     * evaluated.
     *
     * <p>Any exceptions thrown by either {@code test} method are relayed
     * to the caller; if performing first operation throws an exception, the
     * second operation will not be performed.
     *
     * @param p a predicate which will be logically-ANDed with this predicate
     * @return a new predicate which returns {@code true} if either predicate
     * returns {@code true}
     * @throws NullPointerException if p is null
     */
    default BooleanPredicate or(BooleanPredicate p) {
        requireNonNull(p);
        return (value) -> test(value) || p.test(value);
    }
}
