/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */

/**
 * <h1>Square Brackets Java Array Framework</h1>
 *
 * <h2>Overview</h2>
 *
 * <p>Square Brackets is a library for primitive and object arrays leveraging new features of Java 8,
 * such as Lambda and Streams. The library is developed to provide the performance and memory benefits of native
 * (primitive) arrays mixed with the ease of use of the Java Collection Framework. The main features of the framework
 * are:</p>
 *
 * <ul>
 *     <li>Optimized memory footprint:
 *     <ul>
 *         <li>Elements can be stored in primitive arrays, eliminating object overhead;</li>
 *         <li>Elements can be stored in Java NIO buffers, which may reside outside of the normal garbage-collected
 *         heap.</li>
 *     </ul>
 *     </li>
 *     <li>Optimized performance:
 *     <ul>
 *          <li>Constant-time O(1) performance for add, remove, get and set methods;</li>
 *          <li>O(log n) time complexity for element lookup (if elements are sorted);</li>
 *          <li>Primitive iterators and primitive forEach methods, eliminating boxing/unboxing.</li>
 *     </ul>
 *     </li>
 *     <li>Interoperable with Java Collection Framework types (Collections, Lists and Sets):
 *     <ul>
 *         <li>Copy any java.util.Collection into a fixed-size immutable org.squarebrackets.Array and vice-versa;</li>
 *         <li>Wrap any org.squarebrackets.Array in a java.util.List and vice-versa;
 *         <ul>
 *             <li>Access your primitive arrays from a java.util.List view;</li>
 *             <li>Semantics of primitive arrays are identical to semantics of arrays of their object counterparts
 *             (except for Not-a-Number (NaN) values).</li>
 *         </ul>
 *         </li>
 *         <li>Wrap any org.squarebrackets.Array in a live java.util.NavigableSet view;</li>
 *     </ul>
 *     </li>
 *     <li>Promotes defensive programming by enforcing array (im)mutability at compile-time, instead of at runtime. This
 *     is achieved by defining a class hierarchy of immutable, mutable and variable-size arrays.</li>
 * </ul>
 *
 * <h2>Maven</h2>
 *
 * Square Brackets is available from the central maven repository.
 *
 * <pre>
 *{@literal <dependency>}
 *    {@literal <groupId>org.squarebrackets</groupId>}
 *    {@literal <artifactId>squarebrackets</artifactId>}
 *    {@literal <version>1.0.8</version>}
 *{@literal </dependency>}
 * </pre>
 *
 * <h2>Examples</h2>
 *
 * The following section demonstrates the main features of this library.
 *
 * <h3>Immutable / Mutable / Bounded / Unbounded Arrays</h3>
 *
 * <pre>
 * // Contents of immutable array is constant. Array cannot be modified.
 * IntArray immutable = IntArray.copyOf(2, 12, 5, 1, 11, 5, 2, 10, 9, 2);
 * immutable.getInt(0);                    // Returns value at any index.
 * immutable.toArray();                    // Returns a copy as native array.
 * immutable.containsInt(137);             // Tests if value is present in array.
 * immutable.indexOfInt(133);              // Looks up first occurrence of value in array.
 * immutable.lastIndexOfInt(137);          // Looks up last occurrence of value in array.
 * immutable.iterator();                   // Returns a primitive iterator.
 * immutable.forEachInt(                   // Primitive for each method.
 *         System.out::println);
 *
 * // Length of array is constant. Element values can only be replaced,
 * // in addition to immutable array operations.
 * MutableIntArray mutable = MutableIntArray.copyOf(immutable);
 * mutable.setInt(0, 1);                   // Set value at any index.
 * mutable.setAll(array);                  // Replace array values (same length).
 * mutable.replaceAll({@literal v -> v / 2});        // Perform any operation.
 * mutable.sort();                         // Sort array.
 *
 * // Backing store of array is constant. Elements can be added to, and removed
 * // from capacity bounded array in addition to mutable array operations.
 * BoundedIntArray bounded = BoundedIntArray.valueOf(IntBuffer.allocate(10));
 * bounded.addInt(1);                      // Append value.
 * bounded.addAll(array);                  // Append array to array.
 * bounded.setAll(array);                  // Replace array values (length bounded).
 * bounded.intRemove();                    // Remove value.
 * bounded.clear();                        // Remove all values from array.
 *
 * // Array is unbounded, elements can be inserted at, and removed from any location in
 * // addition to bounded array operations. Backing store may be replaced if array
 * // grows or shrinks.
 * UnboundedIntArray unbounded = UnboundedIntArray.copyOf(bounded);
 * unbounded.setAll(array);                // Replace array values (any length).
 * unbounded.addInt(0, 1);                 // Insert value at any index.
 * unbounded.addAll(0, array);             // Insert array at any index.
 * unbounded.removeInt(133);               // Remove value from array if present.
 * unbounded.intRemove(0);                 // Remove value at any index.
 * unbounded.removeIf({@literal v -> v % 2 == 0});   // Remove even values from array.
 * unbounded.distinct();                   // Remove duplicate values from array.
 * </pre>
 *
 * <h3>Streams</h3>
 *
 * <pre>
 * // Subtract 1 and ignore values {@literal >= 10} w/o modifying immutable array using Streams.
 * IntArray array = IntArray.copyOf(2, 12, 5, 1, 11, 5, 2, 10, 9, 2);
 * StringBuilder s = array.intStream().
 *         map({@literal value -> value - 1}).        // No unboxing.
 *         filter({@literal value -> value < 10}).    // No unboxing.
 *         map(Object::toString).
 *         collect(Collectors.toStringBuilder());
 * </pre>
 *
 * <h3>Lambda</h3>
 *
 * <pre>
 * // Subtract 1 and ignore values {@literal >= 10} by modifying and removing values from
 * // unbounded array using Lambda.
 * IntArray array = IntArray.copyOf(2, 12, 5, 1, 11, 5, 2, 10, 9, 2);
 * StringBuilder s = new StringBuilder();
 * UnboundedIntArray unboundedArray = UnboundedIntArray.copyOf(array);
 * unboundedArray.replaceAll({@literal value -> value - 1});    // No unboxing.
 * unboundedArray.removeIf({@literal value -> value >= 10});    // No unboxing.
 * unboundedArray.forEach(s::append);
 * </pre>
 *
 * <h3>Sorting</h3>
 *
 * <pre>
 * UnboundedLongArray array = UnboundedLongArray.newInstance();
 *
 * // Add millions of objects to array: guaranteed constant-time O(1) performance.
 * array.addLong(...);
 * array.addAll(...);
 *
 * array.sort();                    // Sort array: O(n log n) time cost.
 * array.containsLong(value);       // Test presence of value: O(log n) time cost.
 * array.indexOfLong(value);        // Find first occurrence of value: O(log n) time cost.
 * array.lastIndexOfLong(value);    // Find last occurrence of value: O(log n) time cost.
 * </pre>
 *
 * <h3>Distinct</h3>
 *
 * <pre>
 * UnboundedLongArray array = UnboundedLongArray.newInstance();
 *
 * // Add millions of objects to array: guaranteed constant-time O(1) performance.
 * array.addLong(...);
 * array.addAll(...);
 *
 * array.distinct();                // Remove duplicates: O(n) time cost.
 * array.containsLong(value);       // Test presence of value: O(log n) time cost.
 * array.indexOfLong(value);        // Find first occurrence of value: O(log n) time cost.
 * array.lastIndexOfLong(value);    // Find last occurrence of value: O(log n) time cost.
 * </pre>
 *
 * <h3>Sub Arrays</h3>
 *
 * <pre>
 *{@literal UnboundedArray<String> array = UnboundedObjectArray.newInstance();}
 *
 * // Add string objects to array: guaranteed constant-time O(1) performance.
 * array.addLong(...);
 * array.addAll(...);
 *
 * for (String s : array.offset(100)) {
 *     ...                     // Traverse all but first 100 elements.
 * }
 *
 * for (String s : array.length(100)) {
 *     ...                     // Traverse first 100 elements.
 * }
 *
 * // Remove range: constant-time O(1) performance.
 * array.offset(100).clear();  // Remove all but first 100 elements.
 * array.length(100).clear();  // Remove first 100 elements.
 * </pre>
 *
 * <h3>Java Collection Framework Bridge</h3>
 *
 * <pre>
 *{@literal List<Integer> listOfObjects = java.util.Arrays.asList(1, 2, 3);}
 *{@literal Array<Integer> arrayOfObjects = Arrays.newArray(listOfObjects);}
 *
 * IntArray arrayOfPrimitives = IntArray.copyOf(arrayOfObjects);
 *{@literal List<Integer> listOfPrimitives = Arrays.asList(arrayOfPrimitives);}
 *
 * assert arrayOfObjects.equals(arrayOfPrimitives);
 * assert listOfObjects.equals(listOfPrimitives);
 *</pre>
 */
package org.squarebrackets;